from typing import List, Callable

import pyppeteer.page
import pyppeteer.element_handle
from gembox.debug_utils import Debugger

from ._config import PageInteractionConfig
from ..js_util.interface import JsExecutor
from ..js_util.js_handler.action_handler import ClickHandler, InputHandler, ScrollHandler


class PageInteractor(JsExecutor):
    """
    Pyppeteer-based Page Interactor.

    This class is a facade class that provides a high-level interface for interacting with a page.
    """

    def __init__(self, page: pyppeteer.page.Page, config_path: str = None, debug_tool: 'Debugger' = None):
        """
        Initialize the PageInteractor.

        :param page: Pyppeteer page object.
        :param config_path: Path to the interaction configuration. Defaults to None.
        :param debug_tool: Debugging tool. Defaults to Debugger instance.
        """
        # initialize self
        self._page: pyppeteer.page.Page = page
        assert isinstance(self._page, pyppeteer.page.Page), f"page must be an instance of pyppeteer.page.Page, got {type(self._page)}"
        self._config_path = config_path
        self._debug_tool = debug_tool if debug_tool else Debugger()
        self._config: PageInteractionConfig = PageInteractionConfig()
        if self._config_path:
            self._config.parse_config(self._config_path)
        # initialize handlers
        self.click_handler = ClickHandler(page=self._page, js_executor=self, debug_tool=self._debug_tool)
        self.input_handler = InputHandler(page=self._page, js_executor=self, debug_tool=self._debug_tool)
        self.scroll_handler = ScrollHandler(page=self._page, js_executor=self, debug_tool=self._debug_tool)

    async def set_viewport(self, width: int, height: int):
        """
        Set the viewport of the page.

        :param width: (int) Width of the viewport
        :param height: (int) Height of the viewport
        """
        await self._page.setViewport({'width': width, 'height': height})

    async def get_element(self, selector: str) -> pyppeteer.element_handle.ElementHandle:
        return await self._page.querySelector(selector=selector)

    async def get_elements(self, selector: str) -> List[pyppeteer.element_handle.ElementHandle]:
        return await self._page.querySelectorAll(selector=selector)

    # click related
    async def click(self, selector: str, new_page: bool = False):
        """
        Click on an element.

        :param selector: (str) Selector of the element to click
        :param new_page: (bool) Whether to wait for a new page to load, if True, an extra wait time will be added
        :return:
        """
        await self._page.click(selector=selector, new_page=new_page)
        if new_page:
            self._debug_tool.info(f'waiting for new page to load...')
            await self._page.waitFor(self._config.new_page_wait)

    # type related
    async def type_input(self, selector: str, text: str):
        """
        Type text into an input element.

        :param selector: (str) Selector of the element to type into
        :param text: (str) Text to type
        :return:
        """
        return await self.input_handler.type_input(selector=selector, text=text)

    # scroll related
    async def scroll_to_bottom(self):
        """
        Scroll to the bottom of the page.
        """
        return await self.scroll_handler.scroll_to_bottom()

    async def scroll_to_top(self):
        """
        Scroll to the top of the page.
        """
        return await self.scroll_handler.scroll_to_top()

    async def scroll_to(self, x: int, y: int):
        """
        Scroll to a specific position of the page.

        :param x: (int) x coordinate
        :param y: (int) y coordinate
        """
        return await self.scroll_handler.scroll_to(x=x, y=y)

    async def scroll_by(self, x_disp: int, y_disp: int):
        """
        Scroll by a specific displacement.

        :param x_disp: (int) x displacement
        :param y_disp:  (int) y displacement
        :return:
        """
        return await self.scroll_handler.scroll_by(x_disp=x_disp, y_disp=y_disp)

    async def scroll_load(self, scroll_step: int = 400, load_wait: int = 40, same_th: int = 20, scroll_step_callback: Callable = None):
        """
        Scroll and load all contents, until no new content is loaded.

        :param scroll_step: (int) The number of pixels to scroll each time. If None, scroll to bottom.
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param scroll_step_callback: (Callable) A callback function that will be called after each scroll.
        :return:
        """
        return await self.scroll_handler.scroll_load(scroll_step=scroll_step, load_wait=load_wait, same_th=same_th,
                                                     scroll_step_callback=scroll_step_callback)

    async def scroll_load_selector(self, selector: str, threshold: int = None, scroll_step: int = 400,
                                   load_wait: int = 40, same_th: int = 20, scroll_step_callback: Callable = None) \
            -> List[pyppeteer.element_handle.ElementHandle]:
        """
        Scroll and load all contents, until no new content is loaded or enough specific items are collected.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :param scroll_step_callback: (Callable) A callback function that will be called after each scroll.
        :return: (int) The number of elements matching the selector
        """
        return await self.scroll_handler.scroll_load_selector(selector=selector, threshold=threshold,
                                                              scroll_step=scroll_step, load_wait=load_wait,
                                                              same_th=same_th, scroll_step_callback=scroll_step_callback)

    @property
    def url(self) -> str:
        """Return the current url of the page."""
        return self._page.url

    @property
    def config_path(self) -> str:
        """Return the path to the config file."""
        return self._config_path

    @property
    def config(self) -> PageInteractionConfig:
        """Return the config object."""
        return self._config

    async def exec_js(self, js: str):
        """Execute JavaScript code on the page.

        :param js: JavaScript code string to be executed.
        :return: Result of the JavaScript execution.
        """
        return await self._page.evaluate(js)

    def __str__(self):
        return f"PageInteractor(url={self.url})"

    def __repr__(self):
        return self.__str__()


__all__ = ["PageInteractor"]



from functools import wraps

timeout = 5000  # 5 seconds


def wait_for_selector(func):
    @wraps(func)
    async def decorator(self: 'PageInteractor', selector: str, *args, **kwargs):
        await self._page.waitForSelector(selector, timeout=timeout)
        return await func(self, selector, *args, **kwargs)

    return decorator



from zephyrion.pypp.js_util.interface import JsHandler
from zephyrion.pypp.js_util.js_handler.action_handler.common import JsActionHandler
from zephyrion.pypp.page_interactor._decorator import wait_for_selector


class ClickHandler(JsHandler):
    """
    Handler for clicking on elements.
    """
    @wait_for_selector
    async def click(self, selector: str, new_page: bool = False, new_page_wait: float = 1000.):
        """
        Click on an element.

        :param selector: (str) Selector of the element to click
        :param new_page: (bool) Whether to wait for a new page to load, if True, an extra wait time will be added
        :param new_page_wait: (float) Extra wait time in milliseconds for new page to load
        """
        action_handler = JsActionHandler(js_executor=self._js_executor, page=self._page, debug_tool=self.debug_tool)
        self.debug_tool.info(f"Clicking {selector}...")

        await action_handler.click(selector=selector)
        self.debug_tool.info(f'{selector} clicked successfully')
        if new_page:
            self.debug_tool.info(f'waiting for new page to load...')
            await self._page.waitFor(new_page_wait)


__all__ = ['ClickHandler']



class PageInteractionConfig:
    def __init__(self):
        self.new_page_wait = 1000
        """extra waittime when loading new page(in milliseconds)"""
        self.click_pre_wait = 1000
        """time to wait before clicking(in milliseconds)"""
        self.input_pre_wait = 1000
        """time to wait before typing input(in milliseconds)"""
        self.quick_wait = 200
        """quick wait time(in milliseconds)"""
        self.slow_wait = 2000
        """slow wait time(in milliseconds)"""
        self.selector_wait_time_out = 5000
        """selector wait time out(in milliseconds)"""

    def parse_config(self, config_path):
        # Parse config logic
        pass

    def dump_config(self, config_path):
        # Dump config logic
        pass



import abc
import pyppeteer.page

from gembox.debug_utils import Debugger


class JsExecutor(abc.ABC):
    """
    Javascript executor interface.

    Every `JsExecutor` can execute any javascript code.
    """
    @abc.abstractmethod
    async def exec_js(self, js: str):
        """
        Execute javascript code.

        This interface does not define the return value specifications.

        :param js: (str) Javascript code to execute
        :return:
        """
        raise NotImplementedError()


class JsHandler:
    """
    Javascript code handler.

    This class is used to execute javascript code in specific usage case.
    """
    def __init__(self, page: pyppeteer.page.Page, js_executor: JsExecutor, debug_tool=None):
        """
        :param page: (pyppeteer.page.Page) Page to interact with
        :param js_executor: (JsExecutor) Javascript executor to use
        :param debug_tool: (Debugger) Debugger to use
        """
        self._page = page
        self._js_executor = js_executor
        """general javascript code handler"""
        if not isinstance(self._js_executor, JsExecutor):
            raise TypeError(f'js_executor should be an instance of JsExecutor, got {type(js_executor)}')
        self.debug_tool = debug_tool if debug_tool else Debugger()

    @property
    def js_executor(self):
        """
        Javascript executor used by this handler.
        """
        return self._js_executor


__all__ = ['JsExecutor', 'JsHandler']



from functools import wraps


def execute_js(f):
    """
    Decorator for executing javascript code.

    Return RAW javascript result. Usually a `dict` or `str`.
    """
    @wraps(f)
    async def decorator(self, *args, **kwargs):
        js_code = await f(self, *args, **kwargs)
        if js_code:
            result = await self._js_executor.exec_js(js_code)
            self.debug_tool.debug(f"executed js code: {js_code}, result: {result}")
            return result
    return decorator


__all__ = ["execute_js"]



from typing import Union
from functools import wraps

import pyppeteer.page
import pyppeteer.browser
from pyppeteer import launch
from gembox.debug_utils import Debugger


class BrowserNotRunningError(Exception):
    def __init__(self, message=f"Browser is not running"):
        super().__init__(message)


class NoActivePageError(Exception):
    def __init__(self, message=f"No active page found"):
        super().__init__(message)


class NonSingletonError(Exception):
    def __init__(self, message=f"Only one browser instance and one page is allowed"):
        super().__init__(message)


def ensure_browser_is_running(func):
    @wraps(func)
    async def wrapper(browser_mgr, *args, **kwargs):
        if browser_mgr.is_running is False:
            raise BrowserNotRunningError()
        return await func(browser_mgr, *args, **kwargs)

    return wrapper


def ensure_the_page(func):
    @wraps(func)
    async def wrapper(browser_mgr: "SinglePageBrowser", *args, **kwargs):
        the_page = kwargs.pop("page", None)
        # if no `active_page` is provided, get it from `browser_mgr`
        if the_page is None:
            the_page = await browser_mgr.get_page()
        # if active_page is still null, raise error
        if the_page is None:
            raise NoActivePageError()
        return await func(browser_mgr, *args, page=the_page, **kwargs)

    return wrapper


class SinglePageBrowser:
    """
    Browser manager class for managing browser instances.

    To avoid troubles, **only one browser instance** is allowed to run at a time,
    and **only one page** is allowed to be used.
    """

    def __init__(self, browser_options=None, headless=True, debug_tool=None):
        self._is_running: bool = False
        """whether browser is running"""
        self._browser: Union[pyppeteer.browser.Browser, None] = None
        """Singleton browser instance"""
        self._headless: bool = headless
        """whether browser is headless"""
        self._browser_options: dict = browser_options if browser_options is not None else {}
        self._debug_tool = debug_tool if debug_tool is not None else Debugger()

    @property
    def is_running(self) -> bool:
        """whether browser is running"""
        return self._is_running

    @property
    def headless(self) -> bool:
        """whether browser is headless"""
        return self._headless

    async def get_page(self) -> Union[pyppeteer.page.Page, None]:
        """get the active only page, if more than one pages are found, raise error"""
        if not self.is_running:
            return None
        pages = await self._browser.pages()
        if len(pages) == 0:
            return None
        if len(pages) == 1:
            return pages[0]
        raise NonSingletonError(f"Only one page is allowed to be used, but {len(pages)} pages are found")

    async def start_browser(self) -> None:
        page = await self.get_page()
        if page is not None and self._is_running:
            self._debug_tool.warn(f"Browser: Already running.")
        else:
            self._browser = await launch(headless=self._headless, **self._browser_options)
            self._is_running = True

    async def close_browser(self) -> None:
        if not self._is_running or not self._browser:
            self._debug_tool.warn(f"Browser: Not running, no need to close.")
        else:
            await self._browser.close()
            self._browser = None
            self._is_running = False

    async def restart_browser(self) -> None:
        await self.close_browser()
        await self.start_browser()

    @ensure_the_page
    async def get_url(self, page: pyppeteer.page.Page) -> str:
        return page.url

    @ensure_the_page
    async def get_cookies(self, page: pyppeteer.page.Page) -> list[dict[str, Union[str, int, bool]]]:
        return await page.cookies()

    @ensure_the_page
    async def refresh_cookies(self, page: pyppeteer.page.Page):
        """
        Refresh the current page and get the cookies

        :return: (str) cookies
        """
        await page.reload()
        return await self.get_cookies(page=page)

    @ensure_the_page
    async def go_back(self, page: pyppeteer.page.Page) -> None:
        self._debug_tool.info(f"Browser: Go back")
        await page.goBack()

    @ensure_the_page
    async def go(self, url, page: pyppeteer.page.Page) -> None:
        self._debug_tool.info(f"Browser: Go to {url}")
        await page.goto(url)


__all__ = ["SinglePageBrowser"]



class JsGenerator:
    """
    A Basic Javascript code generator.
    """

    # query related
    @staticmethod
    def get_element(selector: str) -> str:
        return f'document.querySelector("{selector}")'

    @staticmethod
    def get_elements(selector: str) -> str:
        return f'document.querySelectorAll("{selector}")'

    @staticmethod
    def get_text_content(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.textContent'

    # attr related
    @staticmethod
    def get_attr(selector: str, attr: str) -> str:
        return f'{JsGenerator.get_element(selector)}.getAttribute("{attr}")'

    @staticmethod
    def set_attr(selector: str, attr: str, value: str) -> str:
        return f'{JsGenerator.get_element(selector)}.setAttribute("{attr}", "{value}")'

    # class list related
    @staticmethod
    def get_class_list(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.classList'

    @staticmethod
    def add_class(selector: str, class_name: str) -> str:
        return f'{JsGenerator.get_class_list(selector)}.add("{class_name}")'

    @staticmethod
    def remove_class(selector: str, class_name: str) -> str:
        return f'{JsGenerator.get_class_list(selector)}.remove("{class_name}")'

    @staticmethod
    def toggle_class(selector: str, class_name: str) -> str:
        return f'{JsGenerator.get_class_list(selector)}.toggle("{class_name}")'

    # action related
    @staticmethod
    def click(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.click()'

    @staticmethod
    def submit(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.submit()'

    @staticmethod
    def focus(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.focus()'

    @staticmethod
    def blur(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.blur()'

    @staticmethod
    def select(selector: str) -> str:
        return f'{JsGenerator.get_element(selector)}.select()'

    # scroll related
    @staticmethod
    def get_scroll_height() -> str:
        return "document.body.scrollHeight"

    @staticmethod
    def get_scroll_width() -> str:
        return "document.body.scrollWidth"

    @staticmethod
    def get_scroll_top() -> str:
        return "window.pageYOffset || document.documentElement.scrollTop"

    @staticmethod
    def get_scroll_left() -> str:
        return "window.pageXOffset || document.documentElement.scrollLeft"

    @staticmethod
    def scroll_to(x: int, y: int):
        return f"window.scrollTo({x}, {y});"

    @staticmethod
    def scroll_by(x_disp: int, y_disp: int):
        return f"window.scrollBy({x_disp}, {y_disp});"

    @staticmethod
    def scroll_to_bottom():
        return f"window.scrollTo(0, {JsGenerator.get_scroll_height()});"

    @staticmethod
    def scroll_to_top():
        return "window.scrollTo(0,0);"


__all__ = ['JsGenerator']



import time
import asyncio
from typing import List, Callable

import pyppeteer.page
import pyppeteer.element_handle

from zephyrion.pypp.js_util.decorator import execute_js
from zephyrion.pypp.js_util.interface import JsHandler, JsExecutor
from zephyrion.pypp.js_util.js_generator import JsGenerator
from zephyrion.pypp.js_util.js_handler.data_handler.common import JsQueryHandler


class ScrollHandler(JsHandler):
    def __init__(self, page: pyppeteer.page.Page, js_executor: JsExecutor, debug_tool=None):
        super().__init__(page=page, js_executor=js_executor, debug_tool=debug_tool)
        self._js_query_handler = JsQueryHandler(page=page, js_executor=js_executor, debug_tool=debug_tool)

    async def scroll_to(self, x: int, y: int):
        return await self._scroll_to(x=x, y=y)

    @execute_js
    async def _scroll_to(self, x: int, y: int):
        return JsGenerator.scroll_to(x=x, y=y)

    async def scroll_by(self, x_disp: int, y_disp: int):
        return await self._scroll_by(x_disp=x_disp, y_disp=y_disp)

    @execute_js
    async def _scroll_by(self, x_disp: int, y_disp: int):
        return JsGenerator.scroll_by(x_disp=x_disp, y_disp=y_disp)

    async def scroll_to_bottom(self):
        return await self._scroll_to_bottom()

    @execute_js
    async def _scroll_to_bottom(self):
        return JsGenerator.scroll_to_bottom()

    async def scroll_to_top(self):
        return await self._scroll_to_top()

    @execute_js
    async def _scroll_to_top(self):
        return JsGenerator.scroll_to_top()

    async def get_scroll_height(self):
        return await self._get_scroll_height()

    @execute_js
    async def _get_scroll_height(self):
        return JsGenerator.get_scroll_height()

    async def get_scroll_width(self):
        return await self._get_scroll_width()

    @execute_js
    async def _get_scroll_width(self):
        return JsGenerator.get_scroll_width()

    async def get_scroll_top(self):
        return await self._get_scroll_top()

    @execute_js
    async def _get_scroll_top(self):
        return JsGenerator.get_scroll_top()

    async def get_scroll_left(self):
        return await self._get_scroll_left()

    @execute_js
    async def _get_scroll_left(self):
        return JsGenerator.get_scroll_left()

    async def _scroll_step(self, scroll_step: int = None) -> None:
        """
        Scroll by `scroll_step` pixels, if scroll_step is `None`, scroll to bottom.
        """
        if scroll_step is None:
            await self.scroll_to_bottom()
        else:
            await self.scroll_by(0, scroll_step)

    async def scroll_load(self, scroll_step: int = 400, load_wait: int = 40, same_th: int = 20, scroll_step_callback: Callable = None):
        """
        Scroll and load all contents, until no new content is loaded.

        :param scroll_step: (int) The number of pixels to scroll each time. If None, scroll to bottom.
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param scroll_step_callback: (Callable) A callback function to be called after each scroll.
        :return:
        """
        return await self._scroll_load_(scroll_step=scroll_step, load_wait=load_wait, same_th=same_th, scroll_step_callback=scroll_step_callback)

    async def scroll_load_selector(self, selector: str, threshold: int = None, scroll_step: int = 400,
                                   load_wait: int = 40, same_th: int = 20, scroll_step_callback: Callable = None) \
            -> List[pyppeteer.element_handle.ElementHandle]:
        """
        Scroll and load all contents, until no new content is loaded or enough specific items are collected.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :param scroll_step_callback: (Callable) A callback function to be called after each scroll.
        :return: (int) The number of elements matching the selector
        """
        self.debug_tool.info(f'Scrolling and loading {selector}...')
        await self._scroll_load_(selector=selector, threshold=threshold, scroll_step=scroll_step, load_wait=load_wait,
                                 same_th=same_th, scroll_step_callback=scroll_step_callback)
        n_elements = await self._js_query_handler.count(selector=selector)
        self.debug_tool.info(f'Loaded {n_elements} elements')
        return await self._js_query_handler.query_all(selector=selector)

    async def _scroll_load_(self, selector: str = None, scroll_step: int = None, load_wait: int = 40,
                            same_th: int = 20, threshold: int = None, scroll_step_callback: Callable = None) -> None:
        """
        Scroll and load all contents.

        It's very common to scroll to the bottom and wait for the page to load until no new content is loaded or enough
        specific items are collected.
        Or you just want to load the whole page.
        This method is to do that.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :return: (None)
        """
        same_count = 0
        last_top = None
        self.debug_tool.info(f'Inner Call Scrolling and loading...(selector={selector}, scroll_step={scroll_step}, load_wait={load_wait}, same_th={same_th}, threshold={threshold})')
        while True:
            if selector is not None:
                count = await self._js_query_handler.count(selector=selector)
                if threshold is not None and count >= threshold:
                    self.debug_tool.info(f'Loaded {count} elements(exceed threshold {threshold}), stop scrolling')
                    break
            await self._scroll_step(scroll_step)
            if scroll_step_callback is not None:
                if asyncio.iscoroutinefunction(scroll_step_callback):
                    await scroll_step_callback()
                else:
                    scroll_step_callback()
            time.sleep(load_wait / 1000.)
            top = await self.get_scroll_top()
            self.debug_tool.debug(f'Scroll top: {top}, last top: {last_top}')
            if top == last_top:
                same_count += 1
                self.debug_tool.info(f'Top unchanged, Scroll top: {top}, last top: {last_top}, same count: {same_count}, same_th: {same_th}')
                if same_count >= same_th:
                    break
            else:
                same_count = 0
            last_top = top


__all__ = ['ScrollHandler']



"""
# Introduction

`js_utils` provides a set of javascript code generators and javascript code handlers.

Including the following modules:

- `js_generator`: javascript code generators, which can generate common javascript code

- `js_handler`: javascript code handlers, which can execute javascript code in specific usage case

# Limitation

`js_generator` can run independently without any dependency.

However, `js_handler` highly depends on  `pyppeteer.page.Page` class, which provide the js execution environment.
Some interaction function is purely based on `pyppeteer.page.Page` class methods for efficiency.
"""


"""
Common Javascript Handlers for actions on elements. *Common* means that other handlers may build upon these.

include:
- `JsQueryHandler`
- `JsAttrHandler`
- `JsClassHandler`
"""
import json

from typing import List
import pyppeteer.element_handle

from zephyrion.pypp.js_util.decorator import execute_js
from zephyrion.pypp.js_util.interface import JsHandler
from zephyrion.pypp.js_util.js_generator import JsGenerator


class JsQueryHandler(JsHandler):
    """
    Javascript Handler for querying elements and pseudo elements.
    """
    async def query_one(self, selector: str) -> pyppeteer.element_handle.ElementHandle:
        """
        Get the first element matching the selector.

        :param selector: (str) Selector of the element to get
        :return: (pyppeteer.element_handle.ElementHandle) Element matching the selector
        """
        return await self._page.querySelector(selector)

    async def query_all(self, selector: str) -> List[pyppeteer.element_handle.ElementHandle]:
        """
        Get all elements matching the selector.

        :param selector: (str) Selector of the element to get
        :return: (list) List of elements matching the selector
        """
        return await self._page.querySelectorAll(selector)

    async def count(self, selector: str) -> int:
        """
        Count the number of elements matching the selector.

        :param selector: (str) Selector of the element to count
        :return: (int) Number of elements matching the selector
        """
        elements = await self.query_all(selector)
        n_elements = 0 if elements is None else len(elements)
        return n_elements

    @execute_js
    async def extract_text(self, selector: str) -> str:
        """
        Get the text content of the first element matching the selector.

        :param selector: (str) Selector of the element to get
        :return: (str) Text content of the first element matching the selector
        """
        return JsGenerator.get_text_content(selector)

    async def extract_texts(self, selector: str) -> List[str]:
        """
        Get the text content of all elements matching the selector.

        :param selector: (str) Selector of the element to get
        :return:
        """

    @execute_js
    async def has_before_pseudo_elements(self, selector: str) -> str:
        js_code_to_exec = f'''
            (function(selector) {{
                const element = document.querySelector(selector);
                if (!element) {{
                    return false;
                }}
                const before = window.getComputedStyle(element, '::before');
                return {{ hasBefore: before.content !== 'none' }};
            }})({json.dumps(selector)})
        '''
        return js_code_to_exec

    @execute_js
    async def has_after_pseudo_elements(self, selector: str) -> str:
        js_code_to_exec = f'''
            (function(selector) {{
                const element = document.querySelector(selector);
                if (!element) {{
                    return false;
                }}
                const after = window.getComputedStyle(element, '::after');
                return {{ hasAfter: after.content !== 'none' }};
            }})({json.dumps(selector)})
        '''
        return js_code_to_exec


class JsAttrHandler(JsHandler):
    """
    Javascript Handler for attribute manipulation.
    """
    async def get_attr(self, selector: str, attr: str):
        """
        Get the value of an attribute of an element.

        :param selector: (str) Selector of the element to get
        :param attr: (str) Attribute to get, e.g. 'href'
        :return: (Any) Value of the attribute
        """
        raw_result = await self._get_attr(selector=selector, attr=attr)
        return raw_result

    @execute_js
    async def _get_attr(self, selector: str, attr: str):
        return JsGenerator.get_attr(selector=selector, attr=attr)

    async def set_attr(self, selector, attr, value) -> None:
        """
        Set the value of an attribute of an element.

        :param selector: (str) Selector of the element to set
        :param attr: (str) Attribute to set, e.g. 'href'
        :param value: (Any) Value to set
        :return: None
        """
        return await self._set_attr(selector=selector, attr=attr, value=value)

    @execute_js
    async def _set_attr(self, selector: str, attr: str, value: str):
        return JsGenerator.set_attr(selector=selector, attr=attr, value=value)


class JsClassHandler(JsHandler):
    """
    Javascript Handler for class list manipulation.
    """
    async def get_class_list(self, selector) -> List[str]:
        """
        Get the class list of an element.

        :param selector: (str) Selector of the element to get
        :return: (list) List of classes
        """
        cls_dict: dict = await self._get_class_list(selector=selector)
        return list(cls_dict.values())

    @execute_js
    async def _get_class_list(self, selector: str):
        return JsGenerator.get_class_list(selector=selector)

    async def add_class(self, selector, class_name) -> None:
        """
        Add a class to an element's classList.

        :param selector: (str) Selector of the element to add
        :param class_name: (str) Class to add
        :return: (None)
        """
        return await self._add_class(selector=selector, class_name=class_name)

    @execute_js
    async def _add_class(self, selector: str, class_name: str):
        return JsGenerator.add_class(selector=selector, class_name=class_name)

    async def remove_class(self, selector, class_name) -> None:
        """
        Remove a class from an element's classList.

        :param selector: (str) Selector of the element to remove
        :param class_name: (str) Class to remove
        :return: (None)
        """
        return await self._remove_class(selector=selector, class_name=class_name)

    @execute_js
    async def _remove_class(self, selector: str, class_name: str):
        return JsGenerator.remove_class(selector=selector, class_name=class_name)

    async def toggle_class(self, selector, class_name) -> None:
        """
        Toggle a class on an element's classList.

        :param selector: (str) Selector of the element to toggle
        :param class_name: (str) Class to toggle
        :return: (None)
        """
        return await self._toggle_class(selector=selector, class_name=class_name)

    @execute_js
    async def _toggle_class(self, selector: str, class_name: str):
        return JsGenerator.toggle_class(selector=selector, class_name=class_name)


__all__ = ["JsClassHandler", "JsAttrHandler", "JsQueryHandler"]



import pathlib
from typing import Union, List, Any

import pyppeteer.element_handle
from gembox.debug_utils import Debugger

from .data_extractor import DataExtractor
from .page_interactor import PageInteractor
from .browser_manager import SinglePageBrowser


class PyppeteerAgent:
    """
    Agent class for managing browser instances and page interactions.
    """
    def __init__(self, headless=False, debug_tool: Debugger = None, interactor_config_path: Union[str, pathlib.Path] = None):
        """
        :param headless: (bool) Whether to run the browser in headless mode
        :param debug_tool: (Debugger) Debugger instance for debugging
        :param interactor_config_path: (str, pathlib.Path) Path to the page interaction config file
        """
        self.debug_tool = debug_tool if debug_tool is not None else Debugger()
        self._interactor_config_path = interactor_config_path
        self.browser_manager = SinglePageBrowser(headless=headless, debug_tool=self.debug_tool)
        self.page_interactor: Union[PageInteractor, None] = None
        self.data_extractor: Union[DataExtractor, None] = None

    @property
    def interactor_config_path(self) -> Union[str, pathlib.Path]:
        """Path to the page interaction config file"""
        return self._interactor_config_path

    @property
    def is_running(self) -> bool:
        """Check if the browser is running"""
        return self.browser_manager.is_running

    @property
    def url(self) -> str:
        """Return the current url of the page."""
        return self.page_interactor.url

    # Context management
    async def start(self) -> None:
        """
        Start the browser and initialize the page interactor and data extractor.

        :return: (None)
        """
        self.debug_tool.debug(f"Starting PyppeteerAgent...")
        await self.browser_manager.start_browser()
        page = await self.browser_manager.get_page()
        self.page_interactor = PageInteractor(page=page, debug_tool=self.debug_tool, config_path=self.interactor_config_path)
        self.data_extractor = DataExtractor(page=page, debug_tool=self.debug_tool)
        assert self.is_running is True, "Browser is not running successfully"
        self.debug_tool.debug(f"Starting PyppeteerAgent Successfully")

    async def stop(self) -> None:
        """
        Close the browser and release the page interactor and data extractor.

        :return: (None)
        """
        await self.browser_manager.close_browser()
        self.page_interactor = None
        self.data_extractor = None
        assert self.is_running is False, "Browser is not closed successfully"

    # Page interactions
    async def click(self, selector: str, new_page: bool = False):
        """
        Click on the element specified by the selector.

        :param selector: (str) Selector for the element to click
        :param new_page: (bool) Whether to open a new page after clicking
        :return: (None)
        """
        return await self.page_interactor.click(selector=selector, new_page=new_page)

    async def type_input(self, selector: str, text: str):
        """
        Type text into an input element.

        :param selector: (str) Selector of the element to type into
        :param text: (str) Text to type
        :return:
        """
        return await self.page_interactor.type_input(selector=selector, text=text)

    async def scroll_to_bottom(self):
        """
        Scroll to the bottom of the page.
        """
        return await self.page_interactor.scroll_to_bottom()

    async def scroll_to_top(self):
        """
        Scroll to the top of the page.
        """
        return await self.page_interactor.scroll_to_top()

    async def scroll_to(self, x: int, y: int):
        """
        Scroll to a specific position of the page.

        :param x: (int) x coordinate
        :param y: (int) y coordinate
        """
        return await self.page_interactor.scroll_to(x=x, y=y)

    async def scroll_by(self, x_disp: int, y_disp: int):
        """
        Scroll by a specific displacement.

        :param x_disp: (int) x displacement
        :param y_disp:  (int) y displacement
        :return:
        """
        return await self.page_interactor.scroll_by(x_disp=x_disp, y_disp=y_disp)

    async def scroll_load(self, scroll_step: int = 400, load_wait: int = 40, same_th: int = 20):
        """
        Scroll and load all contents, until no new content is loaded.

        :param scroll_step: (int) The number of pixels to scroll each time. If None, scroll to bottom.
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :return:
        """
        return await self.page_interactor.scroll_load(scroll_step=scroll_step, load_wait=load_wait, same_th=same_th)

    async def scroll_load_selector(self, selector: str, threshold: int = None, scroll_step: int = 400,
                                   load_wait: int = 40, same_th: int = 20) -> List[pyppeteer.element_handle.ElementHandle]:
        """
        Scroll and load all contents, until no new content is loaded or enough specific items are collected.

        :param selector: (str) The selector of the element to scroll. If None, the method will just scroll to the bottom
        :param scroll_step: (int) The scroll step in pixels. If none, each scroll will be `scroll_to_bottom`
        :param load_wait: (int) The time to wait after each scroll, in milliseconds. If none, the method will wait for 100 ms
        :param same_th: (int) The threshold of the number of same scroll top to stop scrolling.
        :param threshold: (int) only valid when `selector` is not `None`, after loading `threshold` number of elements, the method will stop scrolling
        :return: (int) The number of elements matching the selector
        """
        return await self.page_interactor.scroll_load_selector(selector=selector, threshold=threshold, scroll_step=scroll_step,
                                                               load_wait=load_wait, same_th=same_th)

    # Browser interactions
    async def go_back(self):
        return await self.browser_manager.go_back()

    async def go(self, url: str):
        return await self.browser_manager.go(url=url)

    # data extraction
    async def get_text(self, element: pyppeteer.element_handle.ElementHandle) -> str:
        """
        Extract text from an element.

        :param element: (ElementHandle) Element to extract text from
        :return: (str) Text extracted from the element
        """
        return await self.data_extractor.get_text(element=element)

    async def get_texts(self, selector: str) -> List[str]:
        """
        Extract text from all elements matching the selector.

        :param selector: (str) Selector of the elements to extract text from
        :return: (list) Text extracted from the elements
        """
        return await self.data_extractor.get_texts(selector=selector)

    async def get_attr(self, element: pyppeteer.element_handle.ElementHandle, attribute: str) -> Any:
        """
        Get attribute of given element

        :param element: (ElementHandle) Element to get attribute from
        :param attribute: (str) Attribute to get
        :return: (Any) Attribute of given element
        """
        return await self.data_extractor.get_attr(element=element, attribute=attribute)

    async def get_cls_list(self, element: pyppeteer.element_handle.ElementHandle) -> List[Any]:
        """
        Get classList of given element
        :param element: (ElementHandle) Element to get classList from
        :return: (list) List of classList
        """
        return await self.data_extractor.get_cls_list(element=element)

    async def has_cls(self, element: pyppeteer.element_handle.ElementHandle, cls: str) -> bool:
        """
        Check if given element has given class

        :param element: (ElementHandle) Element to check
        :param cls: (str) Class to check
        :return: (bool) True if element has given class, False otherwise
        """
        return await self.data_extractor.has_cls(element=element, cls=cls)

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()


__all__ = ["PyppeteerAgent"]



"""
Common Javascript Handlers for actions on elements. *Common* means that other handlers may build upon these.

Including:

- `JsActionHandler`
"""
from zephyrion.pypp.js_util.interface import JsHandler
from zephyrion.pypp.js_util.decorator import execute_js
from zephyrion.pypp.js_util.js_generator import JsGenerator


class JsActionHandler(JsHandler):
    """
    Javascript Handler for actions on elements.
    """
    async def click(self, selector):
        return await self._click(selector)

    @execute_js
    async def _click(self, selector: str):
        return JsGenerator.click(selector=selector)

    async def submit(self, selector):
        return await self._submit(selector)

    @execute_js
    async def _submit(self, selector: str):
        return JsGenerator.submit(selector=selector)

    async def focus(self, selector):
        return await self._focus(selector)

    @execute_js
    async def _focus(self, selector: str):
        return JsGenerator.focus(selector=selector)

    async def blur(self, selector):
        return await self._blur(selector)

    @execute_js
    async def _blur(self, selector: str):
        return JsGenerator.blur(selector=selector)

    async def select(self, selector):
        return await self._select(selector)

    @execute_js
    async def _select(self, selector: str):
        return JsGenerator.select(selector=selector)


__all__ = ["JsActionHandler"]



from zephyrion.pypp.js_util.interface import JsHandler
from zephyrion.pypp.page_interactor._decorator import wait_for_selector


class InputHandler(JsHandler):
    """
    Handler for inputting text in elements.
    """
    @wait_for_selector
    async def type_input(self, selector: str, text: str) -> None:
        """
        Type text in an input element.

        :param selector: (str) Selector of the input element
        :param text: (str) Text to type
        :return: (None)
        """
        self.debug_tool.info(f'Typing {text} in {selector}...')
        await self._page.type(selector=selector, text=text)
        self.debug_tool.info(f'{text} typed successfully in {selector}')


__all__ = ['InputHandler']



from typing import List, Any

import pyppeteer.page
from gembox.debug_utils import Debugger
from pyppeteer.element_handle import ElementHandle

from zephyrion.pypp.js_util.js_handler.data_handler.common import JsAttrHandler, JsQueryHandler
from zephyrion.pypp.js_util.interface import JsExecutor


class DataExtractor(JsExecutor):
    """
    Data Extractor for extracting data from elements.
    """
    def __init__(self, page: pyppeteer.page.Page, debug_tool: Debugger = None):
        """
        :param page: (pyppeteer.page.Page) Page to interact with
        :param debug_tool: (Debugger) Debugger to use
        """
        self._page: pyppeteer.page.Page = page
        assert isinstance(self._page, pyppeteer.page.Page)
        self._debug_tool = debug_tool if debug_tool else Debugger()
        self._attr_handler = JsAttrHandler(js_executor=self, page=self._page, debug_tool=self._debug_tool)
        self._query_handler = JsQueryHandler(js_executor=self, page=self._page, debug_tool=self._debug_tool)

    async def get_text(self, element: ElementHandle) -> str:
        """
        Extract text from an element.

        :param element: (ElementHandle) Element to extract text from
        :return: (str) Text extracted from the element
        """
        if not isinstance(element, ElementHandle):
            raise TypeError(f'element should be an instance of ElementHandle, got {type(element)}')
        return await self._page.evaluate('(element) => element.textContent', element)

    async def get_texts(self, selector: str) -> List[str]:
        """
        Extract text from all elements matching the selector.

        :param selector: (str) Selector of the elements to extract text from
        :return: (list) Text extracted from the elements
        """
        elements = await self._page.querySelectorAll(selector)
        return [await self.get_text(element=element) for element in elements]

    async def get_attr(self, element: ElementHandle, attribute: str) -> Any:
        """
        Get attribute of given element

        :param element: (ElementHandle) Element to get attribute from
        :param attribute: (str) Attribute to get
        :return: (Any) Attribute of given element
        """
        return await self._page.evaluate(f'(element) => element.getAttribute("{attribute}")', element)

    async def get_cls_list(self, element: ElementHandle) -> List[Any]:
        """
        Get classList of given element
        :param element: (ElementHandle) Element to get classList from
        :return: (list) List of classList
        """
        cls_dict = await self._page.evaluate('(element) => element.classList', element)
        return list(cls_dict.values())

    async def has_cls(self, element: ElementHandle, cls: str) -> bool:
        """
        Check if given element has given class

        :param element: (ElementHandle) Element to check
        :param cls: (str) Class to check
        :return: (bool) True if element has given class, False otherwise
        """
        cls_list = await self.get_cls_list(element)
        return cls in cls_list

    async def exec_js(self, js: str) -> Any:
        """
        Execute JavaScript code on the page.

        :param js: JavaScript code string to be executed.
        :return: Result of the JavaScript execution.
        """
        return await self._page.evaluate(js)


__all__ = ["DataExtractor"]
